@@ -408,6 +408,193 @@ orc_test_gcc_compile_c64x (OrcProgram *p)
   return ORC_TEST_OK;
 }
 
+OrcTestResult
+orc_test_gcc_compile_riscv (OrcProgram *p)
+{
+  char cmd[400];
+  char *base;
+  char source_filename[100];
+  char obj_filename[100];
+  char dis_filename[100];
+  char dump_filename[100];
+  char dump_dis_filename[100];
+  int ret;
+  FILE *file;
+  OrcCompileResult result;
+  OrcTarget *target;
+  unsigned int flags;
+
+  base = "temp-orc-test";
+
+  sprintf(source_filename, "%s-source.s", base);
+  sprintf(obj_filename, "%s.o", base);
+  sprintf(dis_filename, "%s-source.dis", base);
+  sprintf(dump_filename, "%s-dump.bin", base);
+  sprintf(dump_dis_filename, "%s-dump.dis", base);
+
+  target = orc_target_get_by_name ("riscv");
+  flags = orc_target_get_default_flags (target);
+  flags |= ORC_TARGET_CLEAN_COMPILE;
+
+  result = orc_program_compile_full (p, target, flags);
+  if (!ORC_COMPILE_RESULT_IS_SUCCESSFUL(result)) {
+    /* printf ("  no code generated: %s\n", orc_program_get_error (p)); */
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  fflush (stdout);
+
+  file = fopen (source_filename, "w");
+  fprintf(file, "%s", orc_program_get_asm_code (p));
+  fclose (file);
+
+  sprintf (cmd, "gcc-14 -march=rv64gv_zvkb -Wall "
+      "-c %s -o %s", source_filename, obj_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("riscv gcc failed");
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  sprintf (cmd, "objcopy "
+  "-O binary -j .text "
+  "%s %s", obj_filename, dump_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("riscv objcopy failed");
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  sprintf (cmd, "xxd %s >%s", dump_filename, dis_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("riscv xxd failed");
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  file = fopen (dump_filename, "w");
+  ret = fwrite(p->orccode->code, p->orccode->code_size, 1, file);
+  fclose (file);
+
+  sprintf (cmd, "xxd %s >%s", dump_filename, dump_dis_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("riscv xxd failed");
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  sprintf (cmd, "diff -u %s %s", dis_filename, dump_dis_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    printf("diff failed\n");
+    return ORC_TEST_FAILED;
+  }
+
+  remove (source_filename);
+  remove (obj_filename);
+  remove (dis_filename);
+  remove (dump_filename);
+  remove (dump_dis_filename);
+
+  return ORC_TEST_OK;
+}
+
+OrcTestResult
+orc_test_gcc_compile_loongarch (OrcProgram *p)
+{
+  char cmd[400];
+  char *base;
+  char source_filename[100];
+  char obj_filename[100];
+  char dis_filename[100];
+  char dump_filename[100];
+  char dump_dis_filename[100];
+  int ret;
+  FILE *file;
+  OrcCompileResult result;
+  OrcTarget *target;
+  unsigned int flags;
+
+  base = "temp-orc-test";
+
+  sprintf(source_filename, "%s-source.s", base);
+  sprintf(obj_filename, "%s.o", base);
+  sprintf(dis_filename, "%s-source.dis", base);
+  sprintf(dump_filename, "%s-dump.bin", base);
+  sprintf(dump_dis_filename, "%s-dump.dis", base);
+
+  target = orc_target_get_default ();
+  flags = orc_target_get_default_flags (target);
+  flags |= ORC_TARGET_CLEAN_COMPILE;
+
+  result = orc_program_compile_full (p, target, flags);
+  if (!ORC_COMPILE_RESULT_IS_SUCCESSFUL(result)) {
+    /* printf ("  no code generated: %s\n", orc_program_get_error (p)); */
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  fflush (stdout);
+
+  file = fopen (source_filename, "w");
+  fprintf(file, "%s", orc_program_get_asm_code (p));
+  fclose (file);
+
+  file = fopen (dump_filename, "w");
+  ret = fwrite(p->asm_code, p->orccode->code_size, 1, file);
+  fclose (file);
+
+  sprintf (cmd, "gcc -march=loongarch64 -Wall "
+      "-c %s -o %s", source_filename, obj_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("gcc failed");
+    printf("%s\n", orc_program_get_asm_code (p));
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  sprintf (cmd, "objcopy "
+  "-O binary -j .text "
+  "%s %s", obj_filename, dump_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("objcopy failed");
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  sprintf (cmd, "xxd %s >%s", dump_filename, dis_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("xxd failed");
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  file = fopen (dump_filename, "w");
+  ret = fwrite(p->orccode->code, p->orccode->code_size, 1, file);
+  fclose (file);
+
+  sprintf (cmd, "xxd %s >%s", dump_filename, dump_dis_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    ORC_ERROR ("xxd failed");
+    return ORC_TEST_INDETERMINATE;
+  }
+
+  sprintf (cmd, "diff -u %s %s", dis_filename, dump_dis_filename);
+  ret = system (cmd);
+  if (ret != 0) {
+    printf("diff failed\n");
+    return ORC_TEST_FAILED;
+  }
+
+  remove (source_filename);
+  remove (obj_filename);
+  remove (dis_filename);
+  remove (dump_filename);
+  remove (dump_dis_filename);
+
+  return ORC_TEST_OK;
+}
+
 void
 orc_test_random_bits (void *data, int n_bytes)
 {
@@ -665,16 +852,16 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
 }
 
 static void
-dump_program (const OrcProgram *const program, const OrcTarget *const target)
+dump_program (const OrcProgram *const program, OrcTarget *target)
 {
   char fname[256] = { 0 };
-  snprintf (fname, 256, "%s-%s.S", program->name, target->name);
+  snprintf (fname, 256, "%s-%s.S", program->name, orc_target_get_name (target));
   FILE *f = fopen (fname, "w");
   ORC_ASSERT (f);
   fprintf (f, "%s\n", program->asm_code);
   fclose (f);
 
-  snprintf (fname, 256, "%s-%s.bin", program->name, target->name);
+  snprintf (fname, 256, "%s-%s.bin", program->name, orc_target_get_name (target));
   f = fopen (fname, "wb");
   ORC_ASSERT (f);
   fwrite (program->orccode->code, 1, program->orccode->code_size, f);
