@@ -108,7 +108,7 @@ orc_compiler_powerpc_get_default_flags (void)
 {
   unsigned int flags = 0;
 
-#ifdef __powerpc64__
+#if defined(__powerpc64__) || defined(__ppc64__)
   flags |= ORC_TARGET_POWERPC_64BIT;
 #endif
 #if defined(__LITTLE_ENDIAN__)
@@ -119,8 +119,10 @@ orc_compiler_powerpc_get_default_flags (void)
   flags |= orc_powerpc_cpu_flags;
 #else
   flags |= ORC_TARGET_POWERPC_ALTIVEC;
+#ifndef __APPLE__
   flags |= ORC_TARGET_POWERPC_VSX;
   flags |= ORC_TARGET_POWERPC_V207;
+#endif
 #endif
 
   return flags;
@@ -196,7 +198,7 @@ powerpc_load_constants_outer (OrcCompiler *compiler)
       case ORC_VAR_TYPE_TEMP:
         break;
       default:
-        orc_compiler_error(compiler,"bad vartype");
+        ORC_COMPILER_ERROR(compiler,"bad vartype");
         break;
     }
   }
@@ -265,7 +267,7 @@ orc_powerpc_emit_loop (OrcCompiler* compiler, int update)
 
     ORC_ASM_CODE(compiler,"# %d: %s\n", j, insn->opcode->name);
 
-    compiler->min_temp_reg = ORC_VEC_REG_BASE;
+    orc_compiler_reset_temp_regs (compiler, ORC_VEC_REG_BASE);
 
     compiler->insn_shift = compiler->loop_shift;
     if (insn->flags & ORC_INSTRUCTION_FLAG_X2) {
@@ -279,7 +281,7 @@ orc_powerpc_emit_loop (OrcCompiler* compiler, int update)
     if (rule && rule->emit) {
       rule->emit (compiler, rule->emit_user, insn);
     } else {
-      orc_compiler_error (compiler, "no code generation rule for %s",
+      ORC_COMPILER_ERROR (compiler, "no code generation rule for %s",
           opcode->name);
     }
   }
@@ -345,7 +347,7 @@ orc_compiler_powerpc_assemble (OrcCompiler *compiler)
 
   powerpc_emit_prologue (compiler);
 
-  if (orc_program_has_float (compiler)) {
+  if (orc_compiler_has_float (compiler)) {
     int tmp = POWERPC_V0;
 
     set_vscr = TRUE;
