@@ -5,6 +5,7 @@
 #include <stdlib.h>
 
 #include <orc/orc.h>
+#include <orc/orcinternal.h>
 #include <orc/orcprogram.h>
 #include <orc/orcdebug.h>
 
@@ -89,21 +90,6 @@ orc_compiler_c64x_c_get_default_flags (void)
   return ORC_TARGET_C_NOEXEC;
 }
 
-static const char *varnames[] = {
-  "d1", "d2", "d3", "d4",
-  "s1", "s2", "s3", "s4",
-  "s5", "s6", "s7", "s8",
-  "a1", "a2", "a3", "d4",
-  "c1", "c2", "c3", "c4",
-  "c5", "c6", "c7", "c8",
-  "p1", "p2", "p3", "p4",
-  "p5", "p6", "p7", "p8",
-  "t1", "t2", "t3", "t4",
-  "t5", "t6", "t7", "t8",
-  "t9", "t10", "t11", "t12",
-  "t13", "t14", "t15", "t16"
-};
-
 static void
 output_prototype (OrcCompiler *compiler)
 {
@@ -120,13 +106,13 @@ output_prototype (OrcCompiler *compiler)
       if (need_comma) ORC_ASM_CODE(compiler, ", ");
       if (var->type_name) {
         ORC_ASM_CODE(compiler, "%s * %s", var->type_name,
-            varnames[ORC_VAR_D1 + i]);
+            orc_variable_id_get_name (ORC_VAR_D1 + i));
       } else {
         ORC_ASM_CODE(compiler, "uint%d_t * %s", var->size*8,
-            varnames[ORC_VAR_D1 + i]);
+            orc_variable_id_get_name (ORC_VAR_D1 + i));
       }
       if (p->is_2d) {
-        ORC_ASM_CODE(compiler, ", int %s_stride", varnames[ORC_VAR_D1 + i]);
+        ORC_ASM_CODE(compiler, ", int %s_stride", orc_variable_id_get_name (ORC_VAR_D1 + i));
       }
       need_comma = TRUE;
     }
@@ -137,10 +123,10 @@ output_prototype (OrcCompiler *compiler)
       if (need_comma) ORC_ASM_CODE(compiler, ", ");
       if (var->type_name) {
         ORC_ASM_CODE(compiler, "%s * %s", var->type_name,
-            varnames[ORC_VAR_A1 + i]);
+            orc_variable_id_get_name (ORC_VAR_A1 + i));
       } else {
         ORC_ASM_CODE(compiler, "uint%d_t * %s", var->size*8,
-            varnames[ORC_VAR_A1 + i]);
+            orc_variable_id_get_name (ORC_VAR_A1 + i));
       }
       need_comma = TRUE;
     }
@@ -151,13 +137,13 @@ output_prototype (OrcCompiler *compiler)
       if (need_comma) ORC_ASM_CODE(compiler, ", ");
       if (var->type_name) {
         ORC_ASM_CODE(compiler, "%s * %s", var->type_name,
-            varnames[ORC_VAR_S1 + i]);
+            orc_variable_id_get_name (ORC_VAR_S1 + i));
       } else {
         ORC_ASM_CODE(compiler, "uint%d_t * %s", var->size*8,
-            varnames[ORC_VAR_S1 + i]);
+            orc_variable_id_get_name (ORC_VAR_S1 + i));
       }
       if (p->is_2d) {
-        ORC_ASM_CODE(compiler, ", int %s_stride", varnames[ORC_VAR_S1 + i]);
+        ORC_ASM_CODE(compiler, ", int %s_stride", orc_variable_id_get_name(ORC_VAR_S1 + i));
       }
       need_comma = TRUE;
     }
@@ -166,7 +152,7 @@ output_prototype (OrcCompiler *compiler)
     var = &p->vars[ORC_VAR_P1 + i];
     if (var->size) {
       if (need_comma) ORC_ASM_CODE(compiler, ", ");
-      ORC_ASM_CODE(compiler, "int %s", varnames[ORC_VAR_P1 + i]);
+      ORC_ASM_CODE(compiler, "int %s", orc_variable_id_get_name (ORC_VAR_P1 + i));
       need_comma = TRUE;
     }
   }
@@ -297,7 +283,7 @@ orc_compiler_c64x_c_assemble (OrcCompiler *compiler)
         if (!(compiler->target_flags & ORC_TARGET_C_NOEXEC)) {
           ORC_ASM_CODE(compiler,"  const int var%d = ex->params[%d];\n", i, i);
         } else {
-          ORC_ASM_CODE(compiler,"  const int var%d = %s;\n", i, varnames[i]);
+          ORC_ASM_CODE(compiler,"  const int var%d = %s;\n", i, orc_variable_id_get_name (i));
         }
         break;
       default:
@@ -339,7 +325,7 @@ orc_compiler_c64x_c_assemble (OrcCompiler *compiler)
                 i, i, i);
           } else {
             ORC_ASM_CODE(compiler,"    ptr%d = ORC_PTR_OFFSET(%s, %s_stride * j);\n",
-                i, varnames[i], varnames[i]);
+                i, orc_variable_id_get_name (i), orc_variable_id_get_name (i));
           }
           break;
         case ORC_VAR_TYPE_DEST:
@@ -348,7 +334,7 @@ orc_compiler_c64x_c_assemble (OrcCompiler *compiler)
                 i, i, i);
           } else {
             ORC_ASM_CODE(compiler,"    ptr%d = ORC_PTR_OFFSET(%s, %s_stride * j);\n",
-                i, varnames[i], varnames[i]);
+                i, orc_variable_id_get_name (i), orc_variable_id_get_name (i));
           }
           break;
         default:
@@ -364,14 +350,14 @@ orc_compiler_c64x_c_assemble (OrcCompiler *compiler)
           if (!(compiler->target_flags & ORC_TARGET_C_NOEXEC)) {
             ORC_ASM_CODE(compiler,"  ptr%d = ex->arrays[%d];\n", i, i);
           } else {
-            ORC_ASM_CODE(compiler,"  ptr%d = (void *)%s;\n", i, varnames[i]);
+            ORC_ASM_CODE(compiler,"  ptr%d = (void *)%s;\n", i, orc_variable_id_get_name (i));
           }
           break;
         case ORC_VAR_TYPE_DEST:
           if (!(compiler->target_flags & ORC_TARGET_C_NOEXEC)) {
             ORC_ASM_CODE(compiler,"  ptr%d = ex->arrays[%d];\n", i, i);
           } else {
-            ORC_ASM_CODE(compiler,"  ptr%d = (void *)%s;\n", i, varnames[i]);
+            ORC_ASM_CODE(compiler,"  ptr%d = (void *)%s;\n", i, orc_variable_id_get_name (i));
           }
           break;
         default:
@@ -435,7 +421,7 @@ orc_compiler_c64x_c_assemble (OrcCompiler *compiler)
                 i - ORC_VAR_A1, i);
           } else {
             ORC_ASM_CODE(compiler,"  *%s = (var%d & 0xffff);\n",
-                varnames[i], i);
+                orc_variable_id_get_name (i), i);
           }
         } else {
           if (!(compiler->target_flags & ORC_TARGET_C_NOEXEC)) {
@@ -443,7 +429,7 @@ orc_compiler_c64x_c_assemble (OrcCompiler *compiler)
                 i - ORC_VAR_A1, i);
           } else {
             ORC_ASM_CODE(compiler,"  *%s = var%d;\n",
-                varnames[i], i);
+                orc_variable_id_get_name (i), i);
           }
         }
         break;
