@@ -16,14 +16,27 @@
 #include <orc/orcneon.h>
 
 #define SIZE 65536
+#define MAX_UNROLL 16
+
+enum RegionFlags {
+  FLAG_NUNROLL = (1 << 0),
+  FLAG_REGION1 = (1 << 1),
+  FLAG_REGION2 = (1 << 2),
+  FLAG_REGION3 = (1 << 3),
+};
 
 static void orc_neon_emit_loop (OrcCompiler *compiler, int unroll_index);
 
 extern void orc_compiler_neon_register_rules (OrcTarget *target);
 static unsigned int orc_compiler_neon_get_default_flags (void);
+static enum RegionFlags orc_compiler_neon_calc_regions (const OrcCompiler *compiler);
 
 static void orc_compiler_neon_init (OrcCompiler *compiler);
 static void orc_compiler_neon_assemble (OrcCompiler *compiler);
+static void orc_compiler_neon_assemble_32 (OrcCompiler *compiler);
+static void orc_compiler_neon_assemble_64 (OrcCompiler *compiler);
+
+static void orc_neon64_short_unaligned_loop (OrcCompiler *compiler);
 
 static void orc_neon_save_accumulators (OrcCompiler *compiler);
 static void neon_add_strides (OrcCompiler *compiler);
@@ -157,6 +170,43 @@ orc_compiler_neon_get_default_flags (void)
   return flags;
 }
 
+static enum RegionFlags
+orc_compiler_neon_calc_regions (const OrcCompiler *compiler)
+{
+  enum RegionFlags flags = 0;
+
+  // NOTE: Assume aligned if we know the input size
+  const int constant_n        = compiler->program->constant_n;
+  const orc_bool vectorizable = compiler->loop_shift > 0;
+  const int num_loops         = constant_n >> compiler->loop_shift;
+  const orc_bool n_unrollable = constant_n && (num_loops < MAX_UNROLL);
+  const int elm_per_loop      = 1 << compiler->loop_shift;
+  const int remaining         = constant_n & (elm_per_loop - 1);
+  const orc_bool aligned      = compiler->has_iterator_opcode || (constant_n > 0);
+  const orc_bool enough_n     = constant_n >= elm_per_loop;
+
+  if (!vectorizable)
+    return FLAG_REGION2; // Only region 2
+
+  orc_bool region1 = !aligned;
+  orc_bool region2 = !constant_n || enough_n;
+  orc_bool region3 = !n_unrollable || remaining;
+
+  if (region1)
+    flags |= FLAG_REGION1;
+
+  if (region2)
+    flags |= FLAG_REGION2;
+
+  if (region3)
+    flags |= FLAG_REGION3;
+
+  if (n_unrollable)
+    flags |= FLAG_NUNROLL;
+
+  return flags;
+}
+
 static void
 orc_compiler_neon_init (OrcCompiler *compiler)
 {
@@ -595,9 +645,10 @@ enum {
 };
 
 #define ORC_NEON_ALIGNED_DEST_CUTOFF 64
+#define ORC_NEON_LONG_PROGRAM_CUTOFF 5
 
 static void
-orc_neon64_loop_shift (OrcCompiler *compiler)
+orc_neon64_set_region_counters (OrcCompiler *compiler)
 {
   int align_var;
   int var_size_shift;
@@ -607,51 +658,6 @@ orc_neon64_loop_shift (OrcCompiler *compiler)
   if (compiler->error) return;
   var_size_shift = get_shift (compiler->vars[align_var].size);
 
-  if (compiler->n_insns < 5) {
-    /** Get the number of loops (N) from OrcExecutor */
-    orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_R2, compiler->exec_reg,
-        (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
-
-    /** if N > ORC_NEON_ALIGNED_DEST_CUTOFF, go to LABEL_REGION0_SKIP */
-    orc_arm64_emit_cmp_imm (compiler, 32, ORC_ARM64_R2, ORC_NEON_ALIGNED_DEST_CUTOFF);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_GT, LABEL_REGION0_SKIP);
-
-    /** counter2 = N >> loop shift */
-    orc_arm64_emit_asr_imm (compiler, 32, ORC_ARM64_R1, ORC_ARM64_R2,
-        compiler->loop_shift);
-    orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_R1, compiler->exec_reg,
-        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-
-    /** counter3 = N & loop shift */
-    orc_arm64_emit_and_imm (compiler, 32, ORC_ARM64_R2, ORC_ARM64_R2,
-        (1<<compiler->loop_shift)-1);
-    orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_R2, compiler->exec_reg,
-        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
-
-    /** load function arguments */
-    orc_neon_load_constants_inner (compiler);
-
-    /** if counter2 == zero, go to LABEL_REGION2_SKIP */
-    orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP0, compiler->exec_reg,
-        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-    orc_arm64_emit_cmp_imm (compiler, 32, ORC_ARM64_IP0, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
-
-    /** vector calculation loop */
-    compiler->size_region = 0;
-    orc_arm_emit_label (compiler, LABEL_REGION0_LOOP);
-    orc_arm64_emit_subs_imm (compiler, 32, ORC_ARM64_IP0, ORC_ARM64_IP0, 1);
-
-    /** vector instructions: @todo port to aarch64 */
-    orc_neon_emit_loop (compiler, -1);
-
-    /** if counter2 != zero, repeat loop */
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION0_LOOP);
-    /** else go to LABEL_REGION2_SKIP */
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
-    orc_arm_emit_label (compiler, LABEL_REGION0_SKIP);
-  }
-
   /** IP0 = 1 << align_shift */
   orc_arm64_emit_mov_imm (compiler, 32, ORC_ARM64_IP0, 1<<align_shift);
 
@@ -712,6 +718,53 @@ orc_neon64_loop_shift (OrcCompiler *compiler)
   orc_arm_emit_label (compiler, LABEL_ONE_REGION_AFTER);
 }
 
+static void
+orc_neon64_short_unaligned_loop (OrcCompiler *compiler)
+{
+  /** Get the number of loops (N) from OrcExecutor */
+  orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_R2, compiler->exec_reg,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
+
+  /** if N > ORC_NEON_ALIGNED_DEST_CUTOFF, go to LABEL_REGION0_SKIP */
+  orc_arm64_emit_cmp_imm (compiler, 32, ORC_ARM64_R2, ORC_NEON_ALIGNED_DEST_CUTOFF);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_GT, LABEL_REGION0_SKIP);
+
+  /** counter2 = N >> loop shift */
+  orc_arm64_emit_asr_imm (compiler, 32, ORC_ARM64_R1, ORC_ARM64_R2,
+      compiler->loop_shift);
+  orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_R1, compiler->exec_reg,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+
+  /** counter3 = N & loop shift */
+  orc_arm64_emit_and_imm (compiler, 32, ORC_ARM64_R2, ORC_ARM64_R2,
+      (1<<compiler->loop_shift)-1);
+  orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_R2, compiler->exec_reg,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
+
+  /** load function arguments */
+  orc_neon_load_constants_inner (compiler);
+
+  /** if counter2 == zero, go to LABEL_REGION2_SKIP */
+  orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP0, compiler->exec_reg,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+  orc_arm64_emit_cmp_imm (compiler, 32, ORC_ARM64_IP0, 0);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
+
+  /** vector calculation loop */
+  compiler->size_region = 0;
+  orc_arm_emit_label (compiler, LABEL_REGION0_LOOP);
+  orc_arm64_emit_subs_imm (compiler, 32, ORC_ARM64_IP0, ORC_ARM64_IP0, 1);
+
+  /** vector instructions: @todo port to aarch64 */
+  orc_neon_emit_loop (compiler, -1);
+
+  /** if counter2 != zero, repeat loop */
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION0_LOOP);
+  /** else go to LABEL_REGION2_SKIP */
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
+  orc_arm_emit_label (compiler, LABEL_REGION0_SKIP);
+}
+
 static void
 orc_neon64_loop_caches (OrcCompiler *compiler)
 {
@@ -794,316 +847,10 @@ orc_neon64_loop_caches (OrcCompiler *compiler)
 static void
 orc_compiler_neon_assemble (OrcCompiler *compiler)
 {
-  int align_var;
-  int align_shift;
-  int var_size_shift;
-  int i;
-	int set_fpscr = FALSE;
-
-  align_var = get_align_var (compiler);
-  if (compiler->error) return;
-
-  var_size_shift = get_shift (compiler->vars[align_var].size);
-  align_shift = 4;
-
-  compiler->vars[align_var].is_aligned = FALSE;
-
-  orc_neon_emit_prologue (compiler);
-
-	if (!compiler->is_64bit && orc_program_has_float (compiler)) {
-    set_fpscr = TRUE;
-    ORC_ASM_CODE (compiler,"  vmrs %s, fpscr\n", orc_arm_reg_name (compiler->gp_tmpreg));
-    orc_arm_emit (compiler, 0xeef10a10 | ((compiler->gp_tmpreg&0xf)<<12));
-    ORC_ASM_CODE (compiler,"  push %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
-    orc_arm_emit (compiler, 0xe52d0004 | ((compiler->gp_tmpreg&0xf)<<12));
-
-    orc_arm_emit_load_imm (compiler, compiler->gp_tmpreg, 1<<24);
-    ORC_ASM_CODE (compiler,"  vmsr fpscr, %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
-    orc_arm_emit (compiler, 0xeee10a10 | ((compiler->gp_tmpreg&0xf)<<12));
-  }
-
-  orc_neon_load_constants_outer (compiler);
-
-  if (compiler->is_64bit) {
-    if (compiler->program->is_2d) {
-      if (compiler->program->constant_m > 0) {
-        orc_arm64_emit_mov_imm (compiler, 32, ORC_ARM64_IP1, compiler->program->constant_m);
-        orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
-            (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
-      } else {
-        orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
-            (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A1]));
-        orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
-            (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
-      }
-
-      orc_arm_emit_label (compiler, LABEL_OUTER_LOOP);
-    }
-
-    if (compiler->loop_shift > 0) {
-      orc_neon64_loop_shift (compiler);
-
-      orc_neon_load_constants_inner (compiler);
-
-      orc_neon64_loop_shift_remainder (compiler, counter1,
-          LABEL_REGION1_LOOP, LABEL_REGION1_SKIP);
-      compiler->vars[align_var].is_aligned = TRUE;
-
-      orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP0, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-
-      orc_neon64_loop_caches (compiler);
-
-      compiler->vars[align_var].is_aligned = FALSE;
-      orc_neon64_loop_shift_remainder (compiler, counter3,
-          LABEL_REGION3_LOOP, LABEL_REGION3_SKIP);
-    } else {
-      orc_neon_load_constants_inner (compiler);
-
-      orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP0, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
-
-      orc_neon64_loop_caches (compiler);
-    }
-
-    if (compiler->program->is_2d) {
-      neon_add_strides (compiler);
-
-      orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
-      orc_arm64_emit_subs_imm (compiler, 32, ORC_ARM64_IP1, ORC_ARM64_IP1, 1);
-      orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_OUTER_LOOP);
-    }
-  } else {
-    if (compiler->program->is_2d) {
-      if (compiler->program->constant_m > 0) {
-        orc_arm_emit_load_imm (compiler, ORC_ARM_A3, compiler->program->constant_m);
-        orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-            (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
-      } else {
-        orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-            (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A1]));
-        orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-            (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
-      }
-
-      orc_arm_emit_label (compiler, LABEL_OUTER_LOOP);
-    }
-
-    if (compiler->loop_shift > 0 && compiler->n_insns < 5) {
-      orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
-
-      orc_arm_emit_cmp_imm (compiler, ORC_ARM_A3, ORC_NEON_ALIGNED_DEST_CUTOFF);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_GT, LABEL_REGION0_SKIP);
-
-      orc_arm_emit_asr_imm (compiler, ORC_ARM_A2, ORC_ARM_A3,
-          compiler->loop_shift);
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-
-      orc_arm_emit_and_imm (compiler, ORC_ARM_A3, ORC_ARM_A3,
-          (1<<compiler->loop_shift)-1);
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
-
-      orc_neon_load_constants_inner (compiler);
-      orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-      orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
-
-      compiler->size_region = 0;
-      orc_arm_emit_label (compiler, LABEL_REGION0_LOOP);
-
-      orc_neon_emit_loop (compiler, -1);
-      orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION0_LOOP);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
-      orc_arm_emit_label (compiler, LABEL_REGION0_SKIP);
-    }
-
-    if (compiler->loop_shift > 0) {
-      orc_arm_emit_load_imm (compiler, ORC_ARM_IP, 1<<align_shift);
-
-      orc_arm_emit_load_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,arrays[align_var]));
-      orc_arm_emit_sub (compiler, ORC_ARM_IP, ORC_ARM_IP, ORC_ARM_A2);
-      orc_arm_emit_and_imm (compiler, ORC_ARM_IP, ORC_ARM_IP,
-          (1<<align_shift)-1);
-      if (var_size_shift > 0) {
-        orc_arm_emit_asr_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, var_size_shift);
-      }
-
-      orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
-      orc_arm_emit_cmp (compiler, ORC_ARM_A3, ORC_ARM_IP);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_LE, LABEL_ONE_REGION);
-
-      orc_arm_emit_store_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
-      orc_arm_emit_sub (compiler, ORC_ARM_A2, ORC_ARM_A3, ORC_ARM_IP);
-
-      orc_arm_emit_asr_imm (compiler, ORC_ARM_A3, ORC_ARM_A2,
-          compiler->loop_shift + compiler->unroll_shift);
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-
-      orc_arm_emit_and_imm (compiler, ORC_ARM_A3, ORC_ARM_A2,
-          (1<<(compiler->loop_shift + compiler->unroll_shift))-1);
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
-
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_ONE_REGION_AFTER);
-      orc_arm_emit_label (compiler, LABEL_ONE_REGION);
-
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
-
-      orc_arm_emit_load_imm (compiler, ORC_ARM_A3, 0);
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
-
-      orc_arm_emit_label (compiler, LABEL_ONE_REGION_AFTER);
-    }
-
-    orc_neon_load_constants_inner (compiler);
-
-    if (compiler->loop_shift > 0) {
-      int save_loop_shift = compiler->loop_shift;
-      compiler->loop_shift = 0;
-
-      orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
-
-      orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION1_SKIP);
-
-      orc_arm_emit_label (compiler, LABEL_REGION1_LOOP);
-      orc_neon_emit_loop (compiler, -1);
-      orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION1_LOOP);
-      orc_arm_emit_label (compiler, LABEL_REGION1_SKIP);
-
-      compiler->loop_shift = save_loop_shift;
-      compiler->vars[align_var].is_aligned = TRUE;
-    }
-
-    if (compiler->loop_shift > 0) {
-      orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
-    } else {
-      orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
-    }
-
-    orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
-
-    orc_arm_emit_asr_imm (compiler, compiler->gp_tmpreg, ORC_ARM_IP,
-        17 + var_size_shift - compiler->loop_shift - compiler->unroll_shift);
-    orc_arm_emit_cmp_imm (compiler, compiler->gp_tmpreg, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_MEDIUM);
-
-    /* N is larger than L2 cache size */
-    compiler->size_region = 3;
-    orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_LARGE);
-    for(i=0;i<(1<<compiler->unroll_shift);i++){
-      orc_neon_emit_loop (compiler, i);
-    }
-    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_LARGE);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
-
-    orc_arm_emit_label (compiler, LABEL_REGION2_MEDIUM);
-    orc_arm_emit_asr_imm (compiler, compiler->gp_tmpreg, ORC_ARM_IP,
-        13 + var_size_shift - compiler->loop_shift - compiler->unroll_shift);
-    orc_arm_emit_cmp_imm (compiler, compiler->gp_tmpreg, 0);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SMALL);
-
-    /* N is smaller than L2 cache size */
-    compiler->size_region = 2;
-    orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_MEDIUM);
-    for(i=0;i<(1<<compiler->unroll_shift);i++){
-      orc_neon_emit_loop (compiler, i);
-    }
-    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_MEDIUM);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
-
-    orc_arm_emit_label (compiler, LABEL_REGION2_SMALL);
-    /* N is smaller than L2 cache size */
-    compiler->size_region = 1;
-    orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_SMALL);
-    for(i=0;i<(1<<compiler->unroll_shift);i++){
-      orc_neon_emit_loop (compiler, i);
-    }
-    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
-    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_SMALL);
-
-    orc_arm_emit_label (compiler, LABEL_REGION2_SKIP);
-
-    if (compiler->loop_shift > 0) {
-      int save_loop_shift = compiler->loop_shift;
-
-      compiler->loop_shift = 0;
-
-      compiler->vars[align_var].is_aligned = FALSE;
-
-      orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
-
-      orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION3_SKIP);
-
-      orc_arm_emit_label (compiler, LABEL_REGION3_LOOP);
-      orc_neon_emit_loop (compiler, -1);
-      orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION3_LOOP);
-      orc_arm_emit_label (compiler, LABEL_REGION3_SKIP);
-
-      compiler->loop_shift = save_loop_shift;
-    }
-
-    if (compiler->program->is_2d) {
-      neon_add_strides (compiler);
-
-      orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
-      orc_arm_emit_sub_imm (compiler, ORC_ARM_A3, ORC_ARM_A3, 1, TRUE);
-      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
-          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
-      orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_OUTER_LOOP);
-    }
-  }
-
-  orc_neon_save_accumulators (compiler);
-
-  if (set_fpscr) {
-    ORC_ASM_CODE (compiler,"  pop %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
-    orc_arm_emit (compiler, 0xe49d0004 | ((compiler->gp_tmpreg&0xf)<<12));
-
-    ORC_ASM_CODE (compiler,"  vmsr fpscr, %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
-    orc_arm_emit (compiler, 0xeee10a10 | ((compiler->gp_tmpreg&0xf)<<12));
-  }
-
-  orc_neon_emit_epilogue (compiler);
-
-  orc_arm_emit_align (compiler, 4);
-
-  if (!compiler->is_64bit) {
-    orc_arm_emit_label (compiler, 20);
-    orc_arm_emit_data (compiler, 0x07060706);
-    orc_arm_emit_data (compiler, 0x07060706);
-    orc_arm_emit_data (compiler, 0x0f0e0f0e);
-    orc_arm_emit_data (compiler, 0x0f0e0f0e);
-  }
-
-  orc_arm_do_fixups (compiler);
+  if (compiler->is_64bit)
+    orc_compiler_neon_assemble_64 (compiler);
+  else
+    orc_compiler_neon_assemble_32 (compiler);
 }
 
 static void
@@ -1115,7 +862,8 @@ orc_neon_emit_loop (OrcCompiler *compiler, int unroll_index)
   OrcStaticOpcode *opcode;
   OrcRule *rule;
 
-  orc_compiler_append_code(compiler,"# LOOP shift %d\n", compiler->loop_shift);
+  orc_compiler_append_code(compiler,"# LOOP shift %d (%d)\n",
+      compiler->loop_shift, unroll_index);
   for(j=0;j<compiler->n_insns;j++){
     compiler->insn_index = j;
     insn = compiler->insns + j;
@@ -1226,6 +974,380 @@ orc_neon_emit_loop (OrcCompiler *compiler, int unroll_index)
   }
 }
 
+static void
+orc_compiler_neon_assemble_32 (OrcCompiler *compiler)
+{
+  int align_var;
+  int align_shift;
+  int var_size_shift;
+  int i;
+  int set_fpscr = FALSE;
+
+  align_var = get_align_var (compiler);
+  if (compiler->error) return;
+
+  var_size_shift = get_shift (compiler->vars[align_var].size);
+  align_shift = 4;
+
+  compiler->vars[align_var].is_aligned = FALSE;
+
+  orc_neon_emit_prologue (compiler);
+
+  if (orc_compiler_has_float (compiler)) {
+    set_fpscr = TRUE;
+    ORC_ASM_CODE (compiler,"  vmrs %s, fpscr\n", orc_arm_reg_name (compiler->gp_tmpreg));
+    orc_arm_emit (compiler, 0xeef10a10 | ((compiler->gp_tmpreg&0xf)<<12));
+    ORC_ASM_CODE (compiler,"  push %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
+    orc_arm_emit (compiler, 0xe52d0004 | ((compiler->gp_tmpreg&0xf)<<12));
+
+    orc_arm_emit_load_imm (compiler, compiler->gp_tmpreg, 1<<24);
+    ORC_ASM_CODE (compiler,"  vmsr fpscr, %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
+    orc_arm_emit (compiler, 0xeee10a10 | ((compiler->gp_tmpreg&0xf)<<12));
+  }
+
+  orc_neon_load_constants_outer (compiler);
+
+  if (compiler->program->is_2d) {
+    if (compiler->program->constant_m > 0) {
+      orc_arm_emit_load_imm (compiler, ORC_ARM_A3, compiler->program->constant_m);
+      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    } else {
+      orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A1]));
+      orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    }
+
+    orc_arm_emit_label (compiler, LABEL_OUTER_LOOP);
+  }
+
+  if (compiler->loop_shift > 0 && compiler->n_insns < 5) {
+    orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
+
+    orc_arm_emit_cmp_imm (compiler, ORC_ARM_A3, ORC_NEON_ALIGNED_DEST_CUTOFF);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_GT, LABEL_REGION0_SKIP);
+
+    orc_arm_emit_asr_imm (compiler, ORC_ARM_A2, ORC_ARM_A3,
+        compiler->loop_shift);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+
+    orc_arm_emit_and_imm (compiler, ORC_ARM_A3, ORC_ARM_A3,
+        (1<<compiler->loop_shift)-1);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
+
+    orc_neon_load_constants_inner (compiler);
+    orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+    orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
+
+    compiler->size_region = 0;
+    orc_arm_emit_label (compiler, LABEL_REGION0_LOOP);
+
+    orc_neon_emit_loop (compiler, -1);
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION0_LOOP);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
+    orc_arm_emit_label (compiler, LABEL_REGION0_SKIP);
+  }
+
+  if (compiler->loop_shift > 0) {
+    orc_arm_emit_load_imm (compiler, ORC_ARM_IP, 1<<align_shift);
+
+    orc_arm_emit_load_reg (compiler, ORC_ARM_A2, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,arrays[align_var]));
+    orc_arm_emit_sub (compiler, ORC_ARM_IP, ORC_ARM_IP, ORC_ARM_A2);
+    orc_arm_emit_and_imm (compiler, ORC_ARM_IP, ORC_ARM_IP,
+        (1<<align_shift)-1);
+    if (var_size_shift > 0) {
+      orc_arm_emit_asr_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, var_size_shift);
+    }
+
+    orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
+    orc_arm_emit_cmp (compiler, ORC_ARM_A3, ORC_ARM_IP);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_LE, LABEL_ONE_REGION);
+
+    orc_arm_emit_store_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
+    orc_arm_emit_sub (compiler, ORC_ARM_A2, ORC_ARM_A3, ORC_ARM_IP);
+
+    orc_arm_emit_asr_imm (compiler, ORC_ARM_A3, ORC_ARM_A2,
+        compiler->loop_shift + compiler->unroll_shift);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+
+    orc_arm_emit_and_imm (compiler, ORC_ARM_A3, ORC_ARM_A2,
+        (1<<(compiler->loop_shift + compiler->unroll_shift))-1);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
+
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_ONE_REGION_AFTER);
+    orc_arm_emit_label (compiler, LABEL_ONE_REGION);
+
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
+
+    orc_arm_emit_load_imm (compiler, ORC_ARM_A3, 0);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
+
+    orc_arm_emit_label (compiler, LABEL_ONE_REGION_AFTER);
+  }
+
+  orc_neon_load_constants_inner (compiler);
+
+  if (compiler->loop_shift > 0) {
+    int save_loop_shift = compiler->loop_shift;
+    compiler->loop_shift = 0;
+
+    orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter1));
+
+    orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION1_SKIP);
+
+    orc_arm_emit_label (compiler, LABEL_REGION1_LOOP);
+    orc_neon_emit_loop (compiler, -1);
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION1_LOOP);
+    orc_arm_emit_label (compiler, LABEL_REGION1_SKIP);
+
+    compiler->loop_shift = save_loop_shift;
+    compiler->vars[align_var].is_aligned = TRUE;
+  }
+
+  if (compiler->loop_shift > 0) {
+    orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+  } else {
+    orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
+  }
+
+  orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SKIP);
+
+  orc_arm_emit_asr_imm (compiler, compiler->gp_tmpreg, ORC_ARM_IP,
+      17 + var_size_shift - compiler->loop_shift - compiler->unroll_shift);
+  orc_arm_emit_cmp_imm (compiler, compiler->gp_tmpreg, 0);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_MEDIUM);
+
+  /* N is larger than L2 cache size */
+  compiler->size_region = 3;
+  orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_LARGE);
+  for(i=0;i<(1<<compiler->unroll_shift);i++){
+    orc_neon_emit_loop (compiler, i);
+  }
+  orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_LARGE);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
+
+  orc_arm_emit_label (compiler, LABEL_REGION2_MEDIUM);
+  orc_arm_emit_asr_imm (compiler, compiler->gp_tmpreg, ORC_ARM_IP,
+      13 + var_size_shift - compiler->loop_shift - compiler->unroll_shift);
+  orc_arm_emit_cmp_imm (compiler, compiler->gp_tmpreg, 0);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION2_SMALL);
+
+  /* N is smaller than L2 cache size */
+  compiler->size_region = 2;
+  orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_MEDIUM);
+  for(i=0;i<(1<<compiler->unroll_shift);i++){
+    orc_neon_emit_loop (compiler, i);
+  }
+  orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_MEDIUM);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_AL, LABEL_REGION2_SKIP);
+
+  orc_arm_emit_label (compiler, LABEL_REGION2_SMALL);
+  /* N is smaller than L2 cache size */
+  compiler->size_region = 1;
+  orc_arm_emit_label (compiler, LABEL_REGION2_LOOP_SMALL);
+  for(i=0;i<(1<<compiler->unroll_shift);i++){
+    orc_neon_emit_loop (compiler, i);
+  }
+  orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+  orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION2_LOOP_SMALL);
+
+  orc_arm_emit_label (compiler, LABEL_REGION2_SKIP);
+
+  if (compiler->loop_shift > 0) {
+    int save_loop_shift = compiler->loop_shift;
+
+    compiler->loop_shift = 0;
+
+    compiler->vars[align_var].is_aligned = FALSE;
+
+    orc_arm_emit_load_reg (compiler, ORC_ARM_IP, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3));
+
+    orc_arm_emit_cmp_imm (compiler, ORC_ARM_IP, 0);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_EQ, LABEL_REGION3_SKIP);
+
+    orc_arm_emit_label (compiler, LABEL_REGION3_LOOP);
+    orc_neon_emit_loop (compiler, -1);
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_IP, ORC_ARM_IP, 1, TRUE);
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION3_LOOP);
+    orc_arm_emit_label (compiler, LABEL_REGION3_SKIP);
+
+    compiler->loop_shift = save_loop_shift;
+  }
+
+  if (compiler->program->is_2d) {
+    neon_add_strides (compiler);
+
+    orc_arm_emit_load_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
+    orc_arm_emit_sub_imm (compiler, ORC_ARM_A3, ORC_ARM_A3, 1, TRUE);
+    orc_arm_emit_store_reg (compiler, ORC_ARM_A3, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_OUTER_LOOP);
+  }
+
+  orc_neon_save_accumulators (compiler);
+
+  if (set_fpscr) {
+    ORC_ASM_CODE (compiler,"  pop %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
+    orc_arm_emit (compiler, 0xe49d0004 | ((compiler->gp_tmpreg&0xf)<<12));
+
+    ORC_ASM_CODE (compiler,"  vmsr fpscr, %s\n", orc_arm_reg_name (compiler->gp_tmpreg));
+    orc_arm_emit (compiler, 0xeee10a10 | ((compiler->gp_tmpreg&0xf)<<12));
+  }
+
+  orc_neon_emit_epilogue (compiler);
+
+  orc_arm_emit_align (compiler, 4);
+
+  orc_arm_emit_label (compiler, 20);
+  orc_arm_emit_data (compiler, 0x07060706);
+  orc_arm_emit_data (compiler, 0x07060706);
+  orc_arm_emit_data (compiler, 0x0f0e0f0e);
+  orc_arm_emit_data (compiler, 0x0f0e0f0e);
+
+  orc_arm_do_fixups (compiler);
+}
+
+static void
+orc_compiler_neon_assemble_64 (OrcCompiler *compiler)
+{
+  const enum RegionFlags region_flags = orc_compiler_neon_calc_regions (compiler);
+  const orc_bool region1 = region_flags & FLAG_REGION1;
+  const orc_bool region2 = region_flags & FLAG_REGION2;
+  const orc_bool region3 = region_flags & FLAG_REGION3;
+  const orc_bool nunroll = region_flags & FLAG_NUNROLL;
+  int align_var;
+
+  align_var = get_align_var (compiler);
+  if (compiler->error) return;
+
+  ORC_DEBUG ("Neon compiler regions = [ %s %s %s %s ]"
+      , region1 ? "region1":"-------"
+      , region2 ? "region2":"-------"
+      , region3 ? "region3":"-------"
+      , nunroll ? "nunroll":"-------"
+  );
+
+  compiler->vars[align_var].is_aligned = FALSE;
+
+  orc_neon_emit_prologue (compiler);
+
+  orc_neon_load_constants_outer (compiler);
+
+  if (compiler->program->is_2d) {
+    if (compiler->program->constant_m > 0) {
+      orc_arm64_emit_mov_imm (compiler, 32, ORC_ARM64_IP1, compiler->program->constant_m);
+      orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    } else {
+      orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A1]));
+      orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
+          (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
+    }
+
+    orc_arm_emit_label (compiler, LABEL_OUTER_LOOP);
+  }
+
+  if (!region1 && nunroll) {
+    const int constant_n = compiler->program->constant_n;
+    const int num_loops = constant_n >> compiler->loop_shift;
+    const int remaining = constant_n & ((1 << compiler->loop_shift) - 1);
+    ORC_DEBUG("unrolled loops     = %d", num_loops);
+    ORC_DEBUG("remaining elements = %d", remaining);
+
+    orc_neon_load_constants_inner (compiler);
+
+    for (int i = 0; i < num_loops; i++) {
+      orc_neon_emit_loop (compiler, i);
+    }
+
+    if (region3) {
+      orc_arm64_emit_mov_imm (compiler, ORC_ARM64_REG_32, ORC_ARM64_IP0, remaining);
+      orc_arm_emit_label (compiler, LABEL_REGION3_LOOP);
+      orc_arm64_emit_subs_imm (compiler, ORC_ARM64_REG_32, ORC_ARM64_IP0, ORC_ARM64_IP0, 1);
+
+      int save_loop_shift = compiler->loop_shift; // Ugly, but...uff
+      compiler->loop_shift = 0;
+      orc_neon_emit_loop (compiler, -1);
+      compiler->loop_shift = save_loop_shift;
+
+      orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_REGION3_LOOP);
+    }
+  } else if (compiler->loop_shift > 0) {
+    if (compiler->n_insns < ORC_NEON_LONG_PROGRAM_CUTOFF)
+      orc_neon64_short_unaligned_loop (compiler);
+
+    orc_neon64_set_region_counters (compiler);
+
+    orc_neon_load_constants_inner (compiler);
+
+    orc_neon64_loop_shift_remainder (compiler, counter1,
+        LABEL_REGION1_LOOP, LABEL_REGION1_SKIP);
+    compiler->vars[align_var].is_aligned = TRUE;
+
+    orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP0, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2));
+
+    orc_neon64_loop_caches (compiler);
+
+    compiler->vars[align_var].is_aligned = FALSE;
+    orc_neon64_loop_shift_remainder (compiler, counter3,
+        LABEL_REGION3_LOOP, LABEL_REGION3_SKIP);
+  } else {
+    orc_neon_load_constants_inner (compiler);
+
+    orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP0, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n));
+
+    orc_neon64_loop_caches (compiler);
+  }
+
+  if (compiler->program->is_2d) {
+    neon_add_strides (compiler);
+
+    orc_arm64_emit_load_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor, params[ORC_VAR_A2]));
+    orc_arm64_emit_subs_imm (compiler, 32, ORC_ARM64_IP1, ORC_ARM64_IP1, 1);
+    orc_arm64_emit_store_reg (compiler, 32, ORC_ARM64_IP1, compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A2]));
+    orc_arm_emit_branch (compiler, ORC_ARM_COND_NE, LABEL_OUTER_LOOP);
+  }
+
+  orc_neon_save_accumulators (compiler);
+
+  orc_neon_emit_epilogue (compiler);
+
+  orc_arm_emit_align (compiler, 4);
+
+  orc_arm_do_fixups (compiler);
+}
+
 #define NEON_BINARY(code,a,b,c) \
   ((code) | \
    (((a)&0xf)<<12) | \
