@@ -8,6 +8,7 @@
 #include <orc/orcprogram.h>
 #include <orc/orcutils-private.h>
 #include <orc/orcx86.h>
+#include <orc/orcx86insn.h>
 #include <orc/orcx86-private.h>
 #include <orc/orcinternal.h>
 
@@ -228,22 +229,6 @@ orc_x86_init_accumulators (OrcX86Target *t, OrcCompiler *c)
   }
 }
 
-static void
-orc_x86_load_constant_uint64 (OrcX86Target *t, OrcCompiler *c, int reg, int size,
-    orc_uint64 value)
-{
-  t->load_constant (c, reg, size, value);
-}
-
-static void
-orc_x86_load_constant (OrcCompiler *c, int reg, int size, int value)
-{
-  OrcX86Target *t;
-
-  t = c->target->target_data;
-  orc_x86_load_constant_uint64 (t, c, reg, size, value);
-}
-
 static void
 orc_x86_load_constant_long (OrcX86Target *t, OrcCompiler *c, int reg,
     OrcConstant *constant)
@@ -261,13 +246,8 @@ orc_x86_init_constants (OrcX86Target *t, OrcCompiler *c)
     if (!c->constants[i].alloc_reg)
       continue;
 
-    if (c->constants[i].is_long) {
-      orc_x86_load_constant_long (t, c, c->constants[i].alloc_reg,
-          c->constants + i);
-    } else {
-      orc_x86_load_constant_uint64 (t, c, c->constants[i].alloc_reg, 4,
-          c->constants[i].value);
-    }
+    orc_x86_load_constant_long (t, c, c->constants[i].alloc_reg,
+        c->constants + i);
   }
 }
 
@@ -330,7 +310,7 @@ orc_x86_load_constants_inner (OrcCompiler *c)
       case ORC_VAR_TYPE_TEMP:
         break;
       default:
-        orc_compiler_error (c, "bad vartype");
+        ORC_COMPILER_ERROR (c, "bad vartype");
         break;
     }
   }
@@ -356,7 +336,7 @@ orc_x86_add_strides (OrcCompiler *c)
             c->exec_reg);
 
         if (c->vars[i].ptr_register == 0) {
-          orc_compiler_error (c,
+          ORC_COMPILER_ERROR (c,
               "unimplemented: stride on pointer stored in memory");
         }
         break;
@@ -366,7 +346,7 @@ orc_x86_add_strides (OrcCompiler *c)
       case ORC_VAR_TYPE_TEMP:
         break;
       default:
-        orc_compiler_error (c, "bad vartype");
+        ORC_COMPILER_ERROR (c, "bad vartype");
         break;
     }
   }
@@ -443,7 +423,7 @@ orc_x86_get_max_alignment_var (OrcX86Target *t, OrcCompiler *c)
     return i;
   }
 
-  orc_compiler_error (c, "could not find alignment variable");
+  ORC_COMPILER_ERROR (c, "could not find alignment variable");
   return -1;
 }
 
@@ -485,7 +465,7 @@ orc_x86_emit_split_3_regions (OrcX86Target *t, OrcCompiler *compiler)
   align_shift = var_size_shift + compiler->loop_shift;
 
   /* determine how many iterations until align array is aligned (n1) */
-  orc_x86_emit_mov_imm_reg (compiler, 4, 32, X86_EAX);
+  orc_x86_emit_mov_imm_reg (compiler, 4, t->register_size, X86_EAX);
   // Get the address of the array in question
   // and eax <- eax - addressof(alignment variable)
   orc_x86_emit_sub_memoffset_reg (compiler, 4,
@@ -719,7 +699,8 @@ get_optimised_instruction_order (OrcCompiler *compiler)
 }
 
 static void
-orc_x86_emit_loop (OrcCompiler *compiler, int offset, int update)
+orc_x86_emit_loop (OrcX86Target *t, OrcCompiler *compiler, int offset,
+    int update)
 {
   OrcInstruction *insn;
   OrcStaticOpcode *opcode;
@@ -737,9 +718,8 @@ orc_x86_emit_loop (OrcCompiler *compiler, int offset, int update)
     if (insn->flags & ORC_INSN_FLAG_INVARIANT)
       continue;
 
-    ORC_ASM_CODE (compiler, "# %d: %s\n", j, insn->opcode->name);
-
-    compiler->min_temp_reg = ORC_VEC_REG_BASE;
+    orc_x86_emit_cpuinsn_comment (compiler, "# %d: %s", j, insn->opcode->name);
+    orc_compiler_reset_temp_regs (compiler, t->register_start);
 
     compiler->insn_shift = compiler->loop_shift;
     if (insn->flags & ORC_INSTRUCTION_FLAG_X2) {
@@ -753,7 +733,7 @@ orc_x86_emit_loop (OrcCompiler *compiler, int offset, int update)
     if (rule && rule->emit) {
       rule->emit (compiler, rule->emit_user, insn);
     } else {
-      orc_compiler_error (compiler, "no code generation rule for %s",
+      ORC_COMPILER_ERROR (compiler, "no code generation rule for %s",
           opcode->name);
     }
   }
@@ -812,6 +792,12 @@ orc_x86_clear_emms (OrcX86Target *t, OrcCompiler *c)
     t->clear_emms (c);
 }
 
+static void
+orc_x86_zeroupper (OrcX86Target *t, OrcCompiler *c)
+{
+  if (t->zeroupper)
+    t->zeroupper (c);
+}
 
 static void
 orc_x86_adjust_alignment (OrcX86Target *t, OrcCompiler *compiler)
@@ -832,6 +818,169 @@ orc_x86_adjust_alignment (OrcX86Target *t, OrcCompiler *compiler)
   }
 }
 
+static void
+orc_x86_emit_prologue (OrcCompiler *compiler)
+{
+  orc_compiler_append_code(compiler,".global %s\n", compiler->program->name);
+  orc_compiler_append_code(compiler,".p2align 4\n");
+  orc_compiler_append_code(compiler,"%s:\n", compiler->program->name);
+  if (compiler->is_64bit) {
+    int i;
+
+    orc_x86_emit_cpuinsn_none(compiler, ORC_X86_endbr64);
+    orc_x86_emit_push (compiler, 8, X86_EBP);
+    if (compiler->use_frame_pointer) {
+      orc_x86_emit_mov_reg_reg (compiler, 8, X86_ESP, X86_EBP);
+    }
+
+    for (i = 0; i < 16; i++) {
+      const int reg = ORC_GP_REG_BASE+i;
+      if (compiler->used_regs[reg] &&
+          compiler->save_regs[reg] && reg != X86_EBP) {
+        orc_x86_emit_push (compiler, 8, ORC_GP_REG_BASE+i);
+      }
+    }
+  } else {
+    orc_x86_emit_cpuinsn_none (compiler, ORC_X86_endbr32);
+    orc_x86_emit_push (compiler, 4, X86_EBP);
+    if (compiler->use_frame_pointer) {
+      orc_x86_emit_mov_reg_reg (compiler, 4, X86_ESP, X86_EBP);
+    }
+    orc_x86_emit_mov_memoffset_reg (compiler, 4, 8, X86_ESP, compiler->exec_reg);
+    if (compiler->used_regs[X86_EDI]) {
+      orc_x86_emit_push (compiler, 4, X86_EDI);
+    }
+    if (compiler->used_regs[X86_ESI]) {
+      orc_x86_emit_push (compiler, 4, X86_ESI);
+    }
+    if (compiler->used_regs[X86_EBX]) {
+      orc_x86_emit_push (compiler, 4, X86_EBX);
+    }
+  }
+
+#if 0
+  orc_x86_emit_rdtsc(compiler);
+  orc_x86_emit_mov_reg_memoffset (compiler, 4, X86_EAX,
+      ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A3]), compiler->exec_reg);
+#endif
+}
+
+static void
+orc_x86_emit_epilogue (OrcCompiler *compiler)
+{
+  OrcX86Target *t;
+
+  t = compiler->target->target_data;
+#if 0
+  orc_x86_emit_rdtsc(compiler);
+  orc_x86_emit_mov_reg_memoffset (compiler, 4, X86_EAX,
+      ORC_STRUCT_OFFSET(OrcExecutor,params[ORC_VAR_A4]), compiler->exec_reg);
+#endif
+
+  if (compiler->is_64bit) {
+    int i;
+    for (i = 15; i >= 0; i--) {
+      const int reg = ORC_GP_REG_BASE+i;
+      if (compiler->used_regs[reg] && compiler->save_regs[reg] && reg != X86_EBP) {
+        orc_x86_emit_pop (compiler, 8, reg);
+      }
+    }
+    orc_x86_emit_pop (compiler, 8, X86_EBP);
+  } else {
+    if (compiler->used_regs[X86_EBX]) {
+      orc_x86_emit_pop (compiler, 4, X86_EBX);
+    }
+    if (compiler->used_regs[X86_ESI]) {
+      orc_x86_emit_pop (compiler, 4, X86_ESI);
+    }
+    if (compiler->used_regs[X86_EDI]) {
+      orc_x86_emit_pop (compiler, 4, X86_EDI);
+    }
+    orc_x86_emit_pop (compiler, 4, X86_EBP);
+  }
+  // Remember to yank the higher lanes before returning!
+  // https://stackoverflow.com/a/41349852
+  orc_x86_zeroupper (t, compiler);
+  orc_x86_emit_ret (compiler);
+}
+
+static void
+orc_x86_do_fixups (OrcCompiler *compiler)
+{
+  int i;
+  for(i=0;i<compiler->n_fixups;i++){
+    if (compiler->fixups[i].type == 0) {
+      unsigned char *label = compiler->labels[compiler->fixups[i].label];
+      unsigned char *ptr = compiler->fixups[i].ptr;
+      int diff;
+
+      diff = ((orc_int8)ptr[0]) + (label - ptr);
+      if (diff != (orc_int8)diff) {
+        ORC_COMPILER_ERROR (compiler, "short jump too long %d", diff);
+      }
+
+      ptr[0] = diff;
+    } else if (compiler->fixups[i].type == 1) {
+      unsigned char *label = compiler->labels[compiler->fixups[i].label];
+      unsigned char *ptr = compiler->fixups[i].ptr;
+      int diff;
+
+      diff = ORC_READ_UINT32_LE (ptr) + (label - ptr);
+      ORC_WRITE_UINT32_LE(ptr, diff);
+    }
+  }
+}
+static void
+orc_x86_assemble_copy (OrcCompiler *compiler)
+{
+  OrcInstruction *insn;
+  int shift = 0;
+
+  insn = compiler->program->insns + 0;
+
+  if (strcmp (insn->opcode->name, "copyw") == 0) {
+    shift = 1;
+  } else if (strcmp (insn->opcode->name, "copyl") == 0) {
+    shift = 2;
+  }
+
+  compiler->used_regs[X86_EDI] = TRUE;
+  compiler->used_regs[X86_ESI] = TRUE;
+
+  orc_x86_emit_prologue (compiler);
+
+  orc_x86_emit_mov_memoffset_reg (compiler, 4,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,arrays[insn->dest_args[0]]),
+      compiler->exec_reg, X86_EDI);
+  orc_x86_emit_mov_memoffset_reg (compiler, 4,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,arrays[insn->src_args[0]]),
+      compiler->exec_reg, X86_ESI);
+  orc_x86_emit_mov_memoffset_reg (compiler, 4,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,n), compiler->exec_reg,
+      compiler->gp_tmpreg);
+
+  orc_x86_emit_sar_imm_reg (compiler, 4, 2 - shift, compiler->gp_tmpreg);
+  orc_x86_emit_rep_movs (compiler, 4);
+  if (shift == 0) {
+    orc_x86_emit_mov_memoffset_reg (compiler, 4,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n), compiler->exec_reg,
+        compiler->gp_tmpreg);
+    orc_x86_emit_and_imm_reg (compiler, 4, 3, compiler->gp_tmpreg);
+    orc_x86_emit_rep_movs (compiler, 1);
+  }
+  if (shift == 1) {
+    orc_x86_emit_mov_memoffset_reg (compiler, 4,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n), compiler->exec_reg,
+        compiler->gp_tmpreg);
+    orc_x86_emit_and_imm_reg (compiler, 4, 1, compiler->gp_tmpreg);
+    orc_x86_emit_rep_movs (compiler, 2);
+  }
+
+  orc_x86_emit_epilogue (compiler);
+
+  orc_x86_do_fixups (compiler);
+}
+
 static void
 orc_x86_compile (OrcCompiler *compiler)
 {
@@ -852,7 +1001,7 @@ orc_x86_compile (OrcCompiler *compiler)
 
   is_aligned = compiler->vars[align_var].is_aligned;
   {
-    orc_x86_emit_loop (compiler, 0, 0);
+    orc_x86_emit_loop (t, compiler, 0, 0);
 
     compiler->codeptr = compiler->code;
     free (compiler->asm_code);
@@ -871,7 +1020,7 @@ orc_x86_compile (OrcCompiler *compiler)
 
   orc_x86_save_registers (t, compiler);
 
-  if (t->set_mxcsr && orc_program_has_float (compiler)) {
+  if (t->set_mxcsr && orc_compiler_has_float (compiler)) {
     set_mxcsr = TRUE;
     orc_x86_set_mxcsr (t, compiler);
   }
@@ -930,8 +1079,8 @@ orc_x86_compile (OrcCompiler *compiler)
 
     save_loop_shift = compiler->loop_shift;
     while (n_left >= (1 << compiler->loop_shift)) {
-      ORC_ASM_CODE (compiler, "# AVX LOOP SHIFT %d\n", compiler->loop_shift);
-      orc_x86_emit_loop (compiler, compiler->offset, 0);
+      orc_x86_emit_cpuinsn_comment (compiler, "# AVX LOOP SHIFT %d", compiler->loop_shift);
+      orc_x86_emit_loop (t, compiler, compiler->offset, 0);
 
       n_left -= 1 << compiler->loop_shift;
       compiler->offset += 1 << compiler->loop_shift;
@@ -939,8 +1088,8 @@ orc_x86_compile (OrcCompiler *compiler)
     for (loop_shift = compiler->loop_shift - 1; loop_shift >= 0; loop_shift--) {
       if (n_left >= (1 << loop_shift)) {
         compiler->loop_shift = loop_shift;
-        ORC_ASM_CODE (compiler, "# AVX LOOP SHIFT %d\n", loop_shift);
-        orc_x86_emit_loop (compiler, compiler->offset, 0);
+        orc_x86_emit_cpuinsn_comment (compiler, "# AVX LOOP SHIFT %d", loop_shift);
+        orc_x86_emit_loop (t, compiler, compiler->offset, 0);
         n_left -= 1 << loop_shift;
         compiler->offset += 1 << loop_shift;
       }
@@ -969,12 +1118,12 @@ orc_x86_compile (OrcCompiler *compiler)
 
       for (l = 0; l < save_loop_shift; l++) {
         compiler->loop_shift = l;
-        ORC_ASM_CODE (compiler, "# LOOP SHIFT %d\n", compiler->loop_shift);
+        orc_x86_emit_cpuinsn_comment (compiler, "# LOOP SHIFT %d", compiler->loop_shift);
 
         orc_x86_emit_test_imm_memoffset (compiler, 4, 1 << compiler->loop_shift,
             (int)ORC_STRUCT_OFFSET (OrcExecutor, counter1), compiler->exec_reg);
         orc_x86_emit_je (compiler, LABEL_STEP_UP (compiler->loop_shift));
-        orc_x86_emit_loop (compiler, 0, 1 << compiler->loop_shift);
+        orc_x86_emit_loop (t, compiler, 0, 1 << compiler->loop_shift);
         orc_x86_emit_label (compiler, LABEL_STEP_UP (compiler->loop_shift));
       }
 
@@ -995,7 +1144,7 @@ orc_x86_compile (OrcCompiler *compiler)
           compiler->loop_counter);
     }
 
-    ORC_ASM_CODE (compiler, "# LOOP SHIFT %d\n", compiler->loop_shift);
+    orc_x86_emit_cpuinsn_comment (compiler, "# LOOP SHIFT %d", compiler->loop_shift);
     // Instruction fetch windows are 16-byte aligned
     // https://easyperf.net/blog/2018/01/18/Code_alignment_issues
     orc_x86_emit_align (compiler, 4);
@@ -1003,7 +1152,7 @@ orc_x86_compile (OrcCompiler *compiler)
     ui_max = 1 << compiler->unroll_shift;
     for (ui = 0; ui < ui_max; ui++) {
       compiler->offset = ui << compiler->loop_shift;
-      orc_x86_emit_loop (compiler, compiler->offset,
+      orc_x86_emit_loop (t, compiler, compiler->offset,
           (ui == ui_max - 1)
               << (compiler->loop_shift + compiler->unroll_shift));
     }
@@ -1026,12 +1175,12 @@ orc_x86_compile (OrcCompiler *compiler)
 
       for (l = save_loop_shift - 1; l >= 0; l--) {
         compiler->loop_shift = l;
-        ORC_ASM_CODE (compiler, "# LOOP SHIFT %d\n", compiler->loop_shift);
+        orc_x86_emit_cpuinsn_comment (compiler, "# LOOP SHIFT %d", compiler->loop_shift);
 
         orc_x86_emit_test_imm_memoffset (compiler, 4, 1 << compiler->loop_shift,
             (int)ORC_STRUCT_OFFSET (OrcExecutor, counter3), compiler->exec_reg);
         orc_x86_emit_je (compiler, LABEL_STEP_DOWN (compiler->loop_shift));
-        orc_x86_emit_loop (compiler, 0, 1 << compiler->loop_shift);
+        orc_x86_emit_loop (t, compiler, 0, 1 << compiler->loop_shift);
         orc_x86_emit_label (compiler, LABEL_STEP_DOWN (compiler->loop_shift));
       }
 
@@ -1105,10 +1254,10 @@ orc_x86_register_extension (OrcTarget *t, OrcX86Target *x86t)
   t->get_default_flags = x86t->get_default_flags;
   t->compiler_init = orc_x86_compiler_init;
   t->compile = orc_x86_compile;
-  t->load_constant = orc_x86_load_constant;
   t->get_flag_name = x86t->get_flag_name;
   t->load_constant_long = x86t->load_constant_long;
   t->flush_cache = orc_x86_flush_cache;
   t->target_data = x86t;
+  t->register_size = x86t->register_size;
   orc_target_register (t);
 }
