@@ -6,6 +6,7 @@
 #include <orc/orc.h>
 #include <orc/orcparse.h>
 #include <orc/orcbytecode.h>
+#include <orc/orcutils-private.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -40,7 +41,7 @@ int use_internal = FALSE;
 const char *init_function = NULL;
 const char *decorator = NULL;
 
-char *target = "sse";
+char *target = NULL;
 
 #define ORC_VERSION(a,b,c,d) ((a)*1000000 + (b)*10000 + (c)*100 + (d))
 #define REQUIRE(a,b,c,d) do { \
@@ -460,9 +461,9 @@ main (int argc, char *argv[])
 
       for (i = 0; i < n_programs; i++) {
         char output_file[255] = { 0 };
-        const OrcTarget *const t = orc_target_get_by_name (target);
+        OrcTarget *t = orc_target_get_by_name (target);
         const int l = snprintf (output_file, 255, "%s_%s.bin", programs[i]->name,
-            t->name);
+            orc_target_get_name (t));
         ORC_ASSERT (l > 0);
         FILE *output = fopen (output_file, "wb");
 
